/*
 * Copyright (C) 2009-2015 Typesafe Inc. <http://www.typesafe.com>
 */

import sbt.ScriptedPlugin._
import sbt._
import Keys._
import com.typesafe.tools.mima.plugin.MimaPlugin.mimaDefaultSettings
import com.typesafe.tools.mima.plugin.MimaKeys.{
  previousArtifacts, binaryIssueFilters, reportBinaryIssues
}
import com.typesafe.tools.mima.core._

import com.typesafe.sbt.SbtScalariform.scalariformSettings

import play.twirl.sbt.SbtTwirl
import play.twirl.sbt.Import.TwirlKeys

import sbtdoge.CrossPerProjectPlugin

import bintray.BintrayPlugin.autoImport._

import interplay._
import interplay.Omnidoc.autoImport._
import interplay.PlayBuildBase.autoImport._

import scala.util.control.NonFatal

object BuildSettings {

  // Binary compatibility is tested against this version
  val previousVersions = Set("2.4.0", "2.4.1", "2.4.2", "2.4.3")

  // Argument for setting size of permgen space or meta space for all forked processes
  val maxMetaspace = s"-XX:MaxMetaspaceSize=384m"

  val snapshotBranch = {
    try {
      val branch = "git rev-parse --abbrev-ref HEAD".!!.trim
      if (branch == "HEAD") {
        // not on a branch, get the hash
        "git rev-parse HEAD".!!.trim
      } else branch
    } catch {
      case NonFatal(_) => "unknown"
    }
  }

  /**
   * These settings are used by all projects
   */
  def playCommonSettings: Seq[Setting[_]] = scalariformSettings ++ Seq(
    homepage := Some(url("https://playframework.com")),
    ivyLoggingLevel := UpdateLogging.DownloadOnly,
    resolvers ++= Seq(
      "Scalaz Bintray Repo" at "https://dl.bintray.com/scalaz/releases",
      Resolver.typesafeRepo("releases"),
      Resolver.typesafeIvyRepo("releases")
    ),
    fork in Test := true,
    parallelExecution in Test := false,
    testListeners in (Test,test) := Nil,
    javaOptions in Test ++= Seq(maxMetaspace, "-Xmx512m", "-Xms128m"),
    testOptions += Tests.Argument(TestFrameworks.JUnit, "-v"),
    bintrayPackage := "play-sbt-plugin"
  )

  /**
   * These settings are used by all projects that are part of the runtime, as opposed to development, mode of Play.
   */
  def playRuntimeSettings: Seq[Setting[_]] = playCommonSettings ++ mimaDefaultSettings ++ Seq(
    previousArtifacts := previousVersions.map{ v=>
      if (crossPaths.value) {
        organization.value % s"${moduleName.value}_${scalaBinaryVersion.value}" % v
      } else {
        organization.value % moduleName.value % v
      }
    },
    Docs.apiDocsInclude := true
  )

  def javaVersionSettings(version: String): Seq[Setting[_]] = Seq(
    javacOptions ++= Seq("-source", version, "-target", version),
    javacOptions in doc := Seq("-source", version)
  )

  /**
   * A project that is shared between the SBT runtime and the Play runtime
   */
  def PlayNonCrossBuiltProject(name: String, dir: String): Project = {
    Project(name, file("src/" + dir))
      .enablePlugins(PlaySbtLibrary)
      .settings(playRuntimeSettings: _*)
      .settings(omnidocSettings: _*)
      .settings(
        autoScalaLibrary := false,
        crossPaths := false
      )
  }

  /**
   * A project that is only used when running in development.
   */
  def PlayDevelopmentProject(name: String, dir: String): Project = {
    Project(name, file("src/" + dir))
      .enablePlugins(PlayLibrary)
      .settings(playCommonSettings: _*)
      .settings(
        (javacOptions in compile) ~= (_.map {
          case "1.8" => "1.6"
          case other => other
        })
      )
  }

  /**
   * A project that is in the Play runtime
   */
  def PlayCrossBuiltProject(name: String, dir: String): Project = {
    Project(name, file("src/" + dir))
      .enablePlugins(PlayLibrary)
      .settings(playRuntimeSettings: _*)
      .settings(omnidocSettings: _*)
  }

  def omnidocSettings: Seq[Setting[_]] = Omnidoc.projectSettings ++ Seq(
    omnidocSnapshotBranch := snapshotBranch,
    omnidocPathPrefix := "framework/"
  )

  def playScriptedSettings: Seq[Setting[_]] = Seq(
    ScriptedPlugin.scripted <<= ScriptedPlugin.scripted.tag(Tags.Test),
    scriptedLaunchOpts ++= Seq(
      "-Xmx768m",
      maxMetaspace,
      "-Dscala.version=" + sys.props.get("scripted.scala.version").getOrElse((scalaVersion in PlayBuild.PlayProject).value)
    )
  )

  def playFullScriptedSettings: Seq[Setting[_]] = ScriptedPlugin.scriptedSettings ++ Seq(
    ScriptedPlugin.scriptedLaunchOpts <+= version apply { v => s"-Dproject.version=$v" }
  ) ++ playScriptedSettings

  /**
   * A project that runs in the SBT runtime
   */
  def PlaySbtProject(name: String, dir: String): Project = {
    Project(name, file("src/" + dir))
      .enablePlugins(PlaySbtLibrary)
      .settings(playCommonSettings: _*)
  }

  /**
   * A project that *is* an SBT plugin
   */
  def PlaySbtPluginProject(name: String, dir: String): Project = {
    Project(name, file("src/" + dir))
      .enablePlugins(PlaySbtPlugin)
      .settings(playCommonSettings: _*)
      .settings(playScriptedSettings: _*)
  }

  /**
   * Adds a set of Scala 2.11 modules to the build.
   *
   * Only adds if Scala version is >= 2.11.
   */
  def addScalaModules(modules: ModuleID*): Setting[_] = {
    libraryDependencies := {
      CrossVersion.partialVersion(scalaVersion.value) match {
        case Some((2, scalaMajor)) if scalaMajor >= 11 =>
          libraryDependencies.value ++ modules
        case _ =>
          libraryDependencies.value
      }
    }
  }

  val scalaParserCombinators = "org.scala-lang.modules" %% "scala-parser-combinators" % "1.0.1"
}

object PlayBuild extends Build {

  import Dependencies._
  import BuildSettings._
  import Generators._

  lazy val BuildLinkProject = PlayNonCrossBuiltProject("Build-Link", "build-link")
    .settings(libraryDependencies ++= link)
    .dependsOn(PlayExceptionsProject)

  lazy val RunSupportProject = PlayDevelopmentProject("Run-Support", "run-support")
    .settings(libraryDependencies ++= runSupportDependencies(sbtVersion.value, scalaVersion.value))
    .dependsOn(BuildLinkProject)

  // extra run-support project that is only compiled against sbt scala version
  lazy val SbtRunSupportProject = PlaySbtProject("SBT-Run-Support", "run-support")
    .settings(
      target := target.value / "sbt-run-support",
      libraryDependencies ++= runSupportDependencies(sbtVersion.value, scalaVersion.value)
    ).dependsOn(BuildLinkProject)

  lazy val RoutesCompilerProject = PlayDevelopmentProject("Routes-Compiler", "routes-compiler")
    .enablePlugins(SbtTwirl)
    .settings(
      addScalaModules(scalaParserCombinators),
      libraryDependencies ++= routesCompilerDependencies,
      TwirlKeys.templateFormats := Map("twirl" -> "play.routes.compiler.ScalaFormat")
    )

  lazy val SbtRoutesCompilerProject = PlaySbtProject("SBT-Routes-Compiler", "routes-compiler")
    .enablePlugins(SbtTwirl)
    .settings(
      target := target.value / "sbt-routes-compiler",
      libraryDependencies ++= routesCompilerDependencies,
      TwirlKeys.templateFormats := Map("twirl" -> "play.routes.compiler.ScalaFormat")
    )

  lazy val IterateesProject = PlayCrossBuiltProject("Play-Iteratees", "iteratees")
    .settings(libraryDependencies ++= iterateesDependencies)

  lazy val StreamsProject = PlayCrossBuiltProject("Play-Streams-Experimental", "play-streams")
    .settings(libraryDependencies ++= streamsDependencies)
    .dependsOn(IterateesProject)

  lazy val FunctionalProject = PlayCrossBuiltProject("Play-Functional", "play-functional")

  lazy val DataCommonsProject = PlayCrossBuiltProject("Play-DataCommons", "play-datacommons")

  lazy val JsonProject = PlayCrossBuiltProject("Play-Json", "play-json")
    .settings(
      libraryDependencies ++= jsonDependencies(scalaVersion.value),
      binaryIssueFilters := Seq(
        // These added methods are on sealed traits, so although from a byte code perspective they are binary
        // incompatible, it will only cause issues if someone created a Java class that extends these, because a Scala
        // class wouldn't compile. It's unlikely that anyone created a Java class that extends JsValue or
        // JsLookupResult, if they did, they deserve binary compatibility issues :)
        ProblemFilters.exclude[MissingMethodProblem]("play.api.libs.json.JsLookupResult.validateOpt"),
        ProblemFilters.exclude[MissingMethodProblem]("play.api.libs.json.JsValue.validateOpt")
      )
    )
    .dependsOn(IterateesProject, FunctionalProject, DataCommonsProject)

  lazy val PlayExceptionsProject = PlayNonCrossBuiltProject("Play-Exceptions", "play-exceptions")

  lazy val PlayNettyUtilsProject = PlayNonCrossBuiltProject("Play-Netty-Utils", "play-netty-utils")
    .settings(
      javacOptions in (Compile,doc) += "-Xdoclint:none",
      libraryDependencies ++= nettyUtilsDependencies
    )

  lazy val PlayProject = PlayCrossBuiltProject("Play", "play")
    .enablePlugins(SbtTwirl)
    .settings(
      addScalaModules(scalaParserCombinators),
      libraryDependencies ++= runtime(scalaVersion.value) ++ scalacheckDependencies,
      binaryIssueFilters := Seq(
        // Needed so compile-time DI can use temporary files, raw request bodies
        ProblemFilters.exclude[MissingMethodProblem]("play.api.BuiltInComponents.tempFileCreator")
      ),

      sourceGenerators in Compile <+= (version, scalaVersion, sbtVersion, sourceManaged in Compile) map PlayVersion,

      sourceDirectories in (Compile, TwirlKeys.compileTemplates) := (unmanagedSourceDirectories in Compile).value,
      TwirlKeys.templateImports += "play.api.templates.PlayMagic._",
      mappings in (Compile, packageSrc) ++= {
        // Add both the templates, useful for end users to read, and the Scala sources that they get compiled to,
        // so omnidoc can compile and produce scaladocs for them.
        val twirlSources = (sources in (Compile, TwirlKeys.compileTemplates)).value pair
          relativeTo((sourceDirectories in (Compile, TwirlKeys.compileTemplates)).value)

        val twirlTarget = (target in (Compile, TwirlKeys.compileTemplates)).value
        // The pair with errorIfNone being false both creates the mappings, and filters non twirl outputs out of
        // managed sources
        val twirlCompiledSources = (managedSources in Compile).value.pair(relativeTo(twirlTarget), errorIfNone = false)

        twirlSources ++ twirlCompiledSources
      },
      Docs.apiDocsIncludeManaged := true
    ).settings(Docs.playdocSettings: _*)
     .dependsOn(
      BuildLinkProject,
      IterateesProject % "test->test;compile->compile",
      JsonProject,
      PlayNettyUtilsProject)

  lazy val PlayServerProject = PlayCrossBuiltProject("Play-Server", "play-server")
    .settings(
      libraryDependencies ++= playServerDependencies,
      binaryIssueFilters := Seq(
        ProblemFilters.exclude[MissingClassProblem]("play.core.server.common.ServerRequestUtils"),
        ProblemFilters.exclude[MissingClassProblem]("play.core.server.common.ServerRequestUtils$"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.ForwardedHeaderHandler.play$core$server$common$ForwardedHeaderHandler$$isTrusted"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.ForwardedHeaderHandler.remoteAddress"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.ForwardedHeaderHandler.remoteProtocol"),
        ProblemFilters.exclude[IncompatibleMethTypeProblem]("play.core.server.common.ForwardedHeaderHandler#ForwardedHeaderHandlerConfig.apply"),
        ProblemFilters.exclude[MissingClassProblem]("play.core.server.common.ForwardedHeaderHandler$Version"),
        ProblemFilters.exclude[IncompatibleMethTypeProblem]("play.core.server.common.ForwardedHeaderHandler#ForwardedHeaderHandlerConfig.copy"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.ForwardedHeaderHandler#ForwardedHeaderHandlerConfig.forwardedHeaders"),
        ProblemFilters.exclude[IncompatibleResultTypeProblem]("play.core.server.common.ForwardedHeaderHandler#ForwardedHeaderHandlerConfig.version"),
        ProblemFilters.exclude[IncompatibleMethTypeProblem]("play.core.server.common.ForwardedHeaderHandler#ForwardedHeaderHandlerConfig.this"),
        ProblemFilters.exclude[MissingTypesProblem]("play.core.server.common.NodeIdentifierParser$"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.NodeIdentifierParser.mkList"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.NodeIdentifierParser.failure"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.NodeIdentifierParser.acceptMatch"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.NodeIdentifierParser.acceptIf"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.NodeIdentifierParser.not"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.NodeIdentifierParser.Parser"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.NodeIdentifierParser.literal"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.NodeIdentifierParser.log"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.NodeIdentifierParser.repsep"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.NodeIdentifierParser.acceptSeq"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.NodeIdentifierParser.whiteSpace"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.NodeIdentifierParser.NoSuccess"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.NodeIdentifierParser.success"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.NodeIdentifierParser.rep1"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.NodeIdentifierParser.rep1"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.NodeIdentifierParser.chainr1"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.NodeIdentifierParser.phrase"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.NodeIdentifierParser.opt"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.NodeIdentifierParser.lastNoSuccess"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.NodeIdentifierParser.elem"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.NodeIdentifierParser.elem"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.NodeIdentifierParser.lastNoSuccess_="),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.NodeIdentifierParser.accept"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.NodeIdentifierParser.accept"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.NodeIdentifierParser.accept"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.NodeIdentifierParser.repN"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.NodeIdentifierParser.chainl1"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.NodeIdentifierParser.chainl1"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.NodeIdentifierParser.Failure"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.NodeIdentifierParser.guard"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.NodeIdentifierParser.regex"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.NodeIdentifierParser.Success"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.NodeIdentifierParser.OnceParser"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.NodeIdentifierParser.parse"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.NodeIdentifierParser.parse"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.NodeIdentifierParser.parse"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.NodeIdentifierParser.err"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.NodeIdentifierParser.skipWhitespace"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.NodeIdentifierParser.rep1sep"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.NodeIdentifierParser.parseAll"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.NodeIdentifierParser.parseAll"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.NodeIdentifierParser.parseAll"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.NodeIdentifierParser.Error"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.NodeIdentifierParser.rep"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.NodeIdentifierParser.handleWhiteSpace"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.NodeIdentifierParser.positioned"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.NodeIdentifierParser.~"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.NodeIdentifierParser.parseNode"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.common.NodeIdentifierParser.commit")
      )
    )
    .dependsOn(
      PlayProject,
      IterateesProject % "test->test;compile->compile"
    )

  lazy val PlayNettyServerProject = PlayCrossBuiltProject("Play-Netty-Server", "play-netty-server")
    .settings(
      libraryDependencies ++= netty,
      binaryIssueFilters := Seq(
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.netty.PlayDefaultUpstreamHandler.play$core$server$netty$PlayDefaultUpstreamHandler$$rRemoteAddress$1"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.netty.PlayDefaultUpstreamHandler.play$core$server$netty$PlayDefaultUpstreamHandler$$tryToCreateRequest$1"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.netty.PlayDefaultUpstreamHandler.play$core$server$netty$PlayDefaultUpstreamHandler$$rSecure$1"),
        ProblemFilters.exclude[MissingMethodProblem]("play.core.server.netty.PlayDefaultUpstreamHandler.play$core$server$netty$PlayDefaultUpstreamHandler$$createRequestHeader$1")

      )
    )
    .dependsOn(PlayServerProject)

  // ignored binary incompatibilities for PlayAkkaHttpServerProject since it's an experimental project
  val ignoredABIProblemsPlayAkkaHttpServerProject = {
    import com.typesafe.tools.mima.core._
    import com.typesafe.tools.mima.core.ProblemFilters._
    Seq(
      ProblemFilters.exclude[IncompatibleMethTypeProblem]("play.core.server.akkahttp.ModelConversion.convertRequest"),
      ProblemFilters.exclude[IncompatibleResultTypeProblem]("play.core.server.akkahttp.AkkaHttpServer.materializer"),
      ProblemFilters.exclude[IncompatibleMethTypeProblem]("play.core.server.akkahttp.AkkaStreamsConversion.sourceToEnumerator")
    )
  }

  lazy val PlayAkkaHttpServerProject = PlayCrossBuiltProject("Play-Akka-Http-Server-Experimental", "play-akka-http-server")
    .settings(libraryDependencies ++= akkaHttp)
     // Include scripted tests here as well as in the SBT Plugin, because we
     // don't want the SBT Plugin to have a dependency on an experimental module.
    .settings(playFullScriptedSettings: _*)
    .settings(binaryIssueFilters ++= ignoredABIProblemsPlayAkkaHttpServerProject)
    .dependsOn(PlayServerProject, StreamsProject)
    .dependsOn(PlaySpecs2Project % "test", PlayWsProject % "test")

  lazy val PlayJdbcApiProject = PlayCrossBuiltProject("Play-JDBC-Api", "play-jdbc-api")
    .dependsOn(PlayProject)

  lazy val PlayJdbcProject: Project = PlayCrossBuiltProject("Play-JDBC", "play-jdbc")
    .settings(libraryDependencies ++= jdbcDeps)
    .dependsOn(PlayJdbcApiProject)
    .dependsOn(PlaySpecs2Project % "test")

  lazy val PlayJdbcEvolutionsProject = PlayCrossBuiltProject("Play-JDBC-Evolutions", "play-jdbc-evolutions")
    .dependsOn(PlayJdbcApiProject)
    .dependsOn(PlaySpecs2Project % "test")
    .dependsOn(PlayJdbcProject % "test")
    .dependsOn(PlayJavaJdbcProject % "test")

  lazy val PlayJavaJdbcProject = PlayCrossBuiltProject("Play-Java-JDBC", "play-java-jdbc")
    .settings(libraryDependencies ++= javaJdbcDeps)
    .dependsOn(PlayJdbcProject, PlayJavaProject)
    .dependsOn(PlaySpecs2Project % "test")

  lazy val PlayJpaProject = PlayCrossBuiltProject("Play-Java-JPA", "play-java-jpa")
    .settings(libraryDependencies ++= jpaDeps)
    .dependsOn(PlayJavaJdbcProject)
    .dependsOn(PlayJdbcEvolutionsProject % "test")
    .dependsOn(PlaySpecs2Project % "test")

  lazy val PlayTestProject = PlayCrossBuiltProject("Play-Test", "play-test")
    .settings(
      libraryDependencies ++= testDependencies,
      parallelExecution in Test := false,
      binaryIssueFilters := Seq(
        // play.test.Helpers.invokeWithContext was made static
        ProblemFilters.exclude[MissingMethodProblem]("play.test.Helpers.invokeWithContext")
      )
    ).dependsOn(PlayNettyServerProject)

  lazy val PlaySpecs2Project = PlayCrossBuiltProject("Play-Specs2", "play-specs2")
    .settings(
      libraryDependencies ++= specsBuild,
      parallelExecution in Test := false
    ).dependsOn(PlayTestProject)

  lazy val PlayJavaProject = PlayCrossBuiltProject("Play-Java", "play-java")
    .settings(libraryDependencies ++= javaDeps ++ javaTestDeps)
    .dependsOn(PlayProject % "compile;test->test")

  lazy val PlayDocsProject = PlayCrossBuiltProject("Play-Docs", "play-docs")
    .settings(Docs.settings: _*)
    .settings(
      libraryDependencies ++= playDocsDependencies
    ).dependsOn(PlayNettyServerProject)

  lazy val SbtPluginProject = PlaySbtPluginProject("SBT-Plugin", "sbt-plugin")
    .settings(
      libraryDependencies ++= sbtDependencies(sbtVersion.value, scalaVersion.value),
      sourceGenerators in Compile <+= (version, scalaVersion in PlayProject, sbtVersion, sourceManaged in Compile) map PlayVersion,
      // This only publishes the sbt plugin projects on each scripted run.
      // The runtests script does a full publish before running tests.
      // When developing the sbt plugins, run a publishLocal in the root project first.
      scriptedDependencies := {
        val () = publishLocal.value
        val () = (publishLocal in SbtRoutesCompilerProject).value
      }
    ).dependsOn(SbtRoutesCompilerProject, SbtRunSupportProject)

  val ProtocolCompile = Tags.Tag("protocol-compile")

  val tagProtocolCompileSettings: Seq[Setting[_]] = {
    val settings: Seq[Setting[_]] = Seq(
      compileIncremental <<= compileIncremental tag ProtocolCompile,
      doc <<= doc tag ProtocolCompile
    )
    inConfig(Compile)(settings) ++ inConfig(Test)(settings)
  }

  lazy val ForkRunProtocolProject = PlayDevelopmentProject("Fork-Run-Protocol", "fork-run-protocol")
    .settings(
      libraryDependencies ++= forkRunProtocolDependencies(scalaBinaryVersion.value)
    ).settings(tagProtocolCompileSettings: _*)
    .dependsOn(RunSupportProject)

  // extra fork-run-protocol project that is only compiled against sbt scala version
  lazy val SbtForkRunProtocolProject = PlaySbtProject("SBT-Fork-Run-Protocol", "fork-run-protocol")
    .settings(
      target := target.value / "sbt-fork-run-protocol",
      libraryDependencies ++= forkRunProtocolDependencies(scalaBinaryVersion.value)
    ).settings(tagProtocolCompileSettings: _*)
    .dependsOn(SbtRunSupportProject)

  lazy val ForkRunProject = PlayDevelopmentProject("Fork-Run", "fork-run")
    .settings(
      libraryDependencies ++= forkRunDependencies(scalaBinaryVersion.value),
      // Needed to get the jnotify dependency
      resolvers += Classpaths.sbtPluginReleases
    )
    .dependsOn(ForkRunProtocolProject)

  lazy val SbtForkRunPluginProject = PlaySbtPluginProject("SBT-Fork-Run-Plugin", "sbt-fork-run-plugin")
    .settings(
      libraryDependencies ++= sbtForkRunPluginDependencies(sbtVersion.value, scalaVersion.value),
      // This only publishes the sbt plugin projects on each scripted run.
      // The runtests script does a full publish before running tests.
      // When developing the sbt plugins, run a publishLocal in the root project first.
      scriptedDependencies := {
        val () = publishLocal.value
        val () = (publishLocal in SbtPluginProject).value
        val () = (publishLocal in SbtRoutesCompilerProject).value
      })
    .dependsOn(SbtForkRunProtocolProject, SbtPluginProject)

  lazy val PlayWsProject = PlayCrossBuiltProject("Play-WS", "play-ws")
    .settings(
      libraryDependencies ++= playWsDeps,
      parallelExecution in Test := false,
      // quieten deprecation warnings in tests
      scalacOptions in Test := (scalacOptions in Test).value diff Seq("-deprecation")
    ).dependsOn(PlayProject)
    .dependsOn(PlaySpecs2Project % "test")

  lazy val PlayWsJavaProject = PlayCrossBuiltProject("Play-Java-WS", "play-java-ws")
      .settings(
        libraryDependencies ++= playWsDeps,
        parallelExecution in Test := false
      ).dependsOn(PlayProject)
    .dependsOn(PlayWsProject % "test->test;compile->compile", PlayJavaProject)

  lazy val PlayFiltersHelpersProject = PlayCrossBuiltProject("Filters-Helpers", "play-filters-helpers")
    .settings(
      parallelExecution in Test := false,
      binaryIssueFilters := Seq(
        // AbstractCORSPolicy was never intended to be extended by anything other than Play. It has been made private
        // to reinforce this. The chance that a third party library will have provided an extension to it is very low.
        ProblemFilters.exclude[MissingMethodProblem]("play.filters.cors.AbstractCORSPolicy.errorHandler"),
        // As *Components classes are generally only implemented by end applications, this shouldn't cause a problem.
        // Also, most existing implementations will already have an implementation of this by virtue of the fact that
        // they must implement it for other common *Components interfaces.
        ProblemFilters.exclude[MissingMethodProblem]("play.filters.cors.CORSComponents.httpErrorHandler")
      )
    ).dependsOn(PlayProject, PlaySpecs2Project % "test", PlayJavaProject % "test", PlayWsProject % "test")

  // This project is just for testing Play, not really a public artifact
  lazy val PlayIntegrationTestProject = PlayCrossBuiltProject("Play-Integration-Test", "play-integration-test")
    .settings(
      parallelExecution in Test := false,
      previousArtifacts := Set.empty
    )
    .dependsOn(PlayProject % "test->test", PlayWsProject, PlayWsJavaProject, PlaySpecs2Project)
    .dependsOn(PlayFiltersHelpersProject)
    .dependsOn(PlayJavaProject)
    .dependsOn(PlayAkkaHttpServerProject)

  lazy val PlayCacheProject = PlayCrossBuiltProject("Play-Cache", "play-cache")
    .settings(
      libraryDependencies ++= playCacheDeps,
      parallelExecution in Test := false
    ).dependsOn(PlayProject)
    .dependsOn(PlaySpecs2Project % "test")

  lazy val PlayDocsSbtPlugin = PlaySbtPluginProject("Play-Docs-SBT-Plugin", "play-docs-sbt-plugin")
    .enablePlugins(SbtTwirl)
    .settings(
      libraryDependencies ++= playDocsSbtPluginDependencies
  ).dependsOn(SbtPluginProject)

  lazy val publishedProjects = Seq[ProjectReference](
    PlayProject,
    BuildLinkProject,
    IterateesProject,
    FunctionalProject,
    DataCommonsProject,
    JsonProject,
    RoutesCompilerProject,
    SbtRoutesCompilerProject,
    PlayAkkaHttpServerProject,
    PlayCacheProject,
    PlayJdbcApiProject,
    PlayJdbcProject,
    PlayJdbcEvolutionsProject,
    PlayJavaProject,
    PlayJavaJdbcProject,
    PlayJpaProject,
    PlayNettyUtilsProject,
    PlayNettyServerProject,
    PlayServerProject,
    PlayWsProject,
    PlayWsJavaProject,
    SbtRunSupportProject,
    RunSupportProject,
    SbtPluginProject,
    ForkRunProtocolProject,
    SbtForkRunProtocolProject,
    ForkRunProject,
    SbtForkRunPluginProject,
    PlaySpecs2Project,
    PlayTestProject,
    PlayExceptionsProject,
    PlayDocsProject,
    PlayFiltersHelpersProject,
    PlayIntegrationTestProject,
    PlayDocsSbtPlugin,
    StreamsProject
  )

  lazy val PlayFramework = Project(
    "Play-Framework",
    file("."))
    .enablePlugins(PlayRootProject)
    .enablePlugins(CrossPerProjectPlugin)
    .settings(playCommonSettings: _*)
    .settings(
      playBuildRepoName in ThisBuild := "playframework",
      concurrentRestrictions in Global += Tags.limit(Tags.Test, 1),
      concurrentRestrictions in Global += Tags.limit(ProtocolCompile, 1),
      libraryDependencies ++= (runtime(scalaVersion.value) ++ jdbcDeps),
      Docs.apiDocsInclude := false,
      Docs.apiDocsIncludeManaged := false,
      reportBinaryIssues := (),
      commands += Commands.quickPublish
    ).settings(Release.settings: _*)
    .aggregate(publishedProjects: _*)

}
