/*
 * Copyright (C) 2009-2015 Typesafe Inc. <http://www.typesafe.com>
 */
package play.mvc;

/**
 * Defines a 'call', describing an HTTP request. For example used to create links or populate redirect data.
 * <p>
 * These values are usually generated by the reverse router.
 */
public abstract class Call {

    private static java.util.Random rand = new java.util.Random();

    /**
     * The request URL.
     */
    public abstract String url();

    /**
     * The request HTTP method.
     */
    public abstract String method();

    /**
     * The fragment of the URL.
     */
    public abstract String fragment();

    /**
     * Append a unique identifier to the URL.
     */
    public Call unique() {
        return new play.api.mvc.Call(method(), this.uniquify(this.url()), fragment());
    }

    protected final String uniquify(String url) {
        return url + ((url.indexOf('?') == -1) ? "?" : "&") + rand.nextLong();
    }

    /**
     * Returns a new Call with the given fragment.
     */
    public Call withFragment(String fragment) {
        return new play.api.mvc.Call(method(), url(), fragment);
    }

    /**
     * Returns the fragment (including the leading "#") if this call has one.
     */
    protected String appendFragment() {
        if (this.fragment() != null && !this.fragment().trim().isEmpty()) {
            return "#" + this.fragment();
        } else {
            return "";
        }
    }

    /**
     * Transform this call to an absolute URL.
     */
    public String absoluteURL(Http.Request request) {
        return absoluteURL(request.secure(), request.host());
    }

    /**
     * Transform this call to an absolute URL.
     */
    public String absoluteURL(Http.Request request, boolean secure) {
        return absoluteURL(secure, request.host());
    }

    /**
     * Transform this call to an absolute URL.
     */
    public String absoluteURL(boolean secure, String host) {
        return "http" + (secure ? "s" : "") + "://" + host + this.url() + this.appendFragment();
    }

    /**
     * Transform this call to an WebSocket URL.
     */
    public String webSocketURL(Http.Request request) {
        return webSocketURL(request.secure(), request.host());
    }

    /**
     * Transform this call to an WebSocket URL.
     */
    public String webSocketURL(Http.Request request, boolean secure) {
      return webSocketURL(secure, request.host());
    }

    /**
     * Transform this call to an WebSocket URL.
     */
    public String webSocketURL(boolean secure, String host) {
      return "ws" + (secure ? "s" : "") + "://" + host + this.url();
    }

    public String path() {
        return this.url() + this.appendFragment();
    }

    @Override
    public String toString() {
        return this.path();
    }

}
